-- <<

-------------------------------------------------
--                 GLOBAL VALUES 
-------------------------------------------------

raginature = {}


local difficulty = wesnoth.get_variable("difficulty")
local game_mode  = wesnoth.get_variable("game_mode")
local boss_side_I_died = false
local boss_side_II_died = false

local waves = wesnoth.dofile "~add-ons/raginature_scenario/lua/waves/waves_normal.lua"
local postwave = wesnoth.dofile  "~add-ons/raginature_scenario/lua/waves/postwave_normal.lua"
local postwave_length = 2
local postwave_strength_start = 100
local postwave_strength_increase = 10


local current_wave_nb = 0
local next_wave_turn = 0
local postwave_phase = false 




-------------------------------------------------
--                FIND VACANT TOOL
-------------------------------------------------


-- You can customize your find_vacants function here! 
local width, height, border = wesnoth.get_map_size()
local half = (width - 1) / 2

-- the left area of  the map
local left_area = {
	y_start = border,
	y_end =  height+1,
	x_start = border + half - 1,
	x_end = border - 1
}	

-- the right area of the map
local right_area = {
	y_start = border, 
	y_end =  height+1,
	x_start = border + half + 1,
	x_end = border + width
}


local function set_area_step ( area )
	area.x_step = (area.x_end - area.x_start) / math.abs(area.x_end - area.x_start)
	area.y_step =  (area.y_end - area.y_start) / math.abs(area.y_end - area.y_start)
	return area 
end
	
left_area = set_area_step(left_area)
right_area = set_area_step(right_area)



local helper = wesnoth.require("lua/helper.lua")
local T = wesnoth.require("lua/helper.lua").set_wml_tag_metatable {}

wesnoth.wml_conditionals = wesnoth.wml_conditionals or {}
wesnoth.wml_conditionals.has_unit = wesnoth.wml_conditionals.has_unit or function(cfg)
	raginature.temp = false
	wesnoth.wml_actions["if"] {
		T.have_unit(cfg),
		T["then"] { T.lua { code = "raginature.temp = true" } }
	}
	return raginature.temp
end


local function vacant_filter ( x, y )
	if wesnoth.wml_conditionals.has_unit { x = x, y = y } == false
		and wesnoth.get_terrain(x, y) ~= "Qxu^Xo"
	then
		return { x = x, y = y }
	else 
		return nil
	end
end


local function next_tile ( x, y,  area )
-- go to the next tile
	x = x + area.x_step
	if x == area.x_end then
		x, y = area.x_start, y + area.y_step
	end
	return x, y
end



local function find_vacants ( nb, side, start_pos) 
-- return 'l_vacants': list of {x = <int>, y = <int> }, all the vacants tiles found. If there is no empty tiles enought, return nil
-- return 'next_pos ':  { x = <int>, y = <int> }, the next tile to test for vacant. Use it to increase performance 
-- -- 'nb': integer, the number of vacants to find
-- -- 'side': integer, the human side where the unit should appear
-- -- 'start_pos':  list, { x = <int> , y = <int> }. If not nil, then start at this position instead of x_start and y_start
	local area = side == 1 and left_area or right_area
	local l_vacants = {}
	local nb_found = 0
	local x,y
	if start_pos then 
		x, y = start_pos.x, start_pos.y
	else 
		x, y = area.x_start, area.y_start
	end
	
	while (nb_found < nb) and (y < area.y_end) do
		pos = vacant_filter( x, y )
		if pos then 
			nb_found = nb_found + 1
			table.insert(l_vacants , pos)
		end
		 x, y = next_tile ( x, y, area )
	end
	
	if nb_found < nb then 
		return nil, nil
	else 
		 local next_pos = {x = x, y = y}
		return l_vacants, next_pos
	end
end




-------------------------------------------------
--                FUNCTIONS
-------------------------------------------------



local function side_loose_by_full(side)
-- the side who win by filling of units his opponent's space 
	wesnoth.wml_actions.message {
		speaker = "narrator",
		message = "No free space to place a monster!",
	}
	if game_mode == "coop" then
		wesnoth.wml_actions.kill { side = 1 }
		wesnoth.wml_actions.kill { side = 3 }
	else 
		wesnoth.wml_actions.kill { side = side }
	end
end



local function get_next_wave_turn()
-- return the next wave turn. If no more waves, return nil
	if not postwave_phase then
		local next_wave = waves[current_wave_nb + 1]
		return next_wave and (next_wave.wave_turn )
	else
		return (wesnoth.current.turn + postwave_length)
	end
end


local function get_ai_facing(side)
-- get the ai facing from the human side 'side'
	if side == 1 then return "sw"
	elseif side == 3 then return "se"
	else return nil
	end
end


local function set_label_nextwave(turn)
-- update the next wave label with a new value turn
	wesnoth.wml_actions.label {
		x = 12, y = 13,
		text = string.format("<span color='#FFFFFF'>Next wave:\n    turn %s</span>", turn)
	}
	wesnoth.wml_actions.label {
		x = 12, y = 4,
		text = string.format("<span color='#FFFFFF'>Next wave:\n    turn %s</span>", turn)
	}
end


local function clear_label_nextwave()
	wesnoth.wml_actions.label {x = 12, y = 13, text = ""}
	wesnoth.wml_actions.label {x = 12, y = 4, text = ""}
end


local function set_elite_trait(unit) 
-- set elite trait to unit for unpoisonable, unslowable and unplagueable
	local elite_ability = { "elite", { id = "elite", name = "elite", description="This unit is an elite. It cannot be poisoned, slowed or turned into a walking corpse. However, attacks with the special poison, slow or plague inflict 25% more damage."}}
	wesnoth.add_modification(unit, "trait",
		{
			{"effect" , {apply_to="status",add="unpoisonable"}},
			{"effect" , {apply_to="status",add="unplagueable"}},
			{"effect" , {apply_to="new_ability",{"abilities", {elite_ability}}}},
			{"effect" , {apply_to="halo",halo="elite_crown.png"}},
		}
	)
end


local function set_unit_strength(unit, strength)
	local increase = strength - 100
	local strength_ability = { "strength", {
		id = "strength", 
		name = "strength +" .. increase .. "%", 
		description="The strength of this unit increase more and more. +" .. increase .. "% more hitpoints, +" .. increase .. "% more damage"
	}}
	wesnoth.add_modification(unit, "trait", {
		{"effect", { apply_to = "attack", increase_damage = increase .. "%" } },
		{"effect", { apply_to = "hitpoints", increase_total = increase .. "%", heal_full = true } },
		{"effect", { apply_to = "new_ability", {"abilities", { strength_ability } } } },
	})
	return unit 
end


local function final_boss_wave(msg, img)
	wesnoth.play_sound("boss_is_coming.ogg")
	wesnoth.music_list.clear()
	wesnoth.music_list.add("silence.ogg", true)
	wesnoth.music_list.add("into_the_shadows.ogg")
	wesnoth.music_list.add("frantic.ogg")
	wesnoth.music_list.add("northerners.ogg")
	wesnoth.music_list.add("weight_of_revenge.ogg")
	clear_label_nextwave()
	wesnoth.wml_actions.message{ speaker = "narrator", message = msg, image = img }
end



local function create_fake_mob(side, nb, unit_type, elite, strength, id, start_pos)
-- create 'nb' mobs or  elites of type 'unit_type' on the human side 'side'
-- -- 'elite': <boolean>, If it's an elite, set it true, otherise, false.
-- -- 'strength': <integer>, if there is a strength modification (in percent), otherise, set nil 
-- -- 'id': <string>, id of mob, only usefull for final boss, otherise, set nil
	local l_vacants, last_pos = find_vacants (nb, side, start_pos)
	if not l_vacants then 
		side_loose_by_full(side)
		return nil
	else
		for i=1,nb do
			local pos = l_vacants[i]
			local mob = wesnoth.create_unit { type=unit_type, side=side+1, facing=get_ai_facing(side) }
			mob.status.not_real = true
			mob.status.unhealable = true
			mob.status.invulnerable = true
			mob.attacks_left = 0
			mob.moves = 0
			if elite == true then 
				set_elite_trait(mob) 
				mob.variables.elite = true
			end
			if strength then 
				set_unit_strength(mob, strength)
				mob.variables.strength = strength
			end
			if id then 
				mob.id = id
			end
			wesnoth.add_modification(mob, "object",{id="raginature_mob_gs", { "effect", {apply_to="image_mod", replace="GS( )"}}})
			wesnoth.put_unit(mob, pos.x, pos.y)
		end
		return last_pos
	end 
end



local function create_fake_mob_both_sides(nb, unit_type, elite, strength, pos_I, pos_II)
-- create mobs on both sides. If game_mode == "solo" then create only on side 2
	if game_mode ~= "solo" then
		pos_I = create_fake_mob(1, nb, unit_type, elite, strength, nil, pos_I)
		pos_II = create_fake_mob(3, nb, unit_type, elite, strength, nil, pos_II)
	else
		pos_I = create_fake_mob(1, nb, unit_type, elite, strength, nil, pos_I)
	end
	return pos_I, pos_II
end



local function create_wave(wave, strength)
-- create every units from the 'wave' array. Use 'strength' if there is a strength modification (percent), otherise, set nil 
-- -- 'pos_I', 'pos_II': list {x = <int>, y = <int> }, contain the position next the last created monster. II for side 2, III for side 3
	local pos_I, pos_II
	for _, unit in ipairs(wave) do
		local rank = unit[1]
		
		if rank == "mob" then 
			pos_I, pos_II = create_fake_mob_both_sides(unit.number, unit.unit_type, false, strength, pos_I, pos_II)
			
		elseif rank == "elite" then
			pos_I, pos_II = create_fake_mob_both_sides(unit.number, unit.unit_type, true, strength, pos_I, pos_II)
			
		elseif rank == "boss" then
			if unit.final == false then
				wesnoth.wml_actions.message{ speaker = "narrator", message = unit.message, image = unit.image }
				pos_I, pos_II = create_fake_mob_both_sides( 1, unit.unit_type, true, strength, pos_I, pos_II)
				
			else 
				final_boss_wave(unit.message, unit.image)
				pos_I = create_fake_mob(1, 1, unit.unit_type, true, strength, "boss_side_II", pos_I)
				if game_mode ~= "solo" then
					pos_II = create_fake_mob(3, 1, unit.unit_type, true, strength, "boss_side_III", pos_II)
				end
			end
	
		end
	end
end


local function create_wave_by_nb(wave_nb)
-- create a wave by its number, its position in the waves file 'waves'
	local wave = waves[wave_nb]
	create_wave(wave, nil)
end


local function make_real_mobs(side)
	local not_real_mobs = wesnoth.get_units { side = side, status="not_real" }
	for _, mob in ipairs(not_real_mobs) do
		local mob_copy = wesnoth.create_unit { id=mob.id, type=mob.type, side=side, facing= mob.facing, x= mob.x, y=mob.y}
		if mob.variables.elite == true then 
			set_elite_trait(mob_copy)
		end
		if mob.variables.strength then 
			set_unit_strength(mob_copy, mob.variables.strength)
		end 
		wesnoth.erase_unit(mob)
		wesnoth.put_unit(mob_copy)
	end 
end


local function postwave_phase_init()
	postwave_phase = true
	current_wave_nb = 0
	next_wave_turn = wesnoth.current.turn + postwave_length
	set_label_nextwave(next_wave_turn)
	wesnoth.wml_actions.label {
		x = 12, y = 12,
		text = string.format("<span color='#FF0000'>   The wind has stopped,\n animals have disappeared</span>")
	}
	wesnoth.wml_actions.label {
		x = 12, y = 3,
		text = string.format("<span color='#FF0000'>   The wind has stopped,\n animals have disappeared</span>")
	}
	return next_wave_turn
end




-------------------------------------------------
--                   EVENTS
-------------------------------------------------


wesnoth.wml_actions.event {
	name = "die",
	{"filter", { id="boss_side_II" } },
	{"lua", { code = "raginature.boss_side_I_die()" } },
}


wesnoth.wml_actions.event {
	name = "die",
	{"filter", { id="boss_side_III" } },
	{"lua", { code = "raginature.boss_side_II_die()" } },
}


wesnoth.wml_actions.event {
	name = "side turn end",
	first_time_only=false,
	{"lua", { code = "raginature.side_turn_end_event()" } },
}


wesnoth.wml_actions.event {
	name = "turn refresh",
	first_time_only=false,
	{"lua", { code = "raginature.turn_refresh()" } },
}





-------------------------------------------------
--                       MAIN CALLS 
-------------------------------------------------


-- kill unnecessary leaders
wesnoth.wml_actions.kill {canrecruit = true, side = 2 , fire_event = false, animate = false } 
wesnoth.wml_actions.kill {canrecruit = true, side = 4 , fire_event = false, animate = false } 
if game_mode == "solo" then 
	wesnoth.wml_actions.kill {canrecruit = true, side = 3, fire_event = false, animate = false } 
end
	

local next_wave_turn = get_next_wave_turn()

set_label_nextwave(next_wave_turn)



-- If one (at least) final boss die during a duel mode, then init the postwave phase!
-- If the 2 final boss die during a coop-mode, then ending with victory!
-- If the boss 
function raginature.boss_side_I_die()
	boss_side_I_died = true
	if game_mode == "coop" then 
		if boss_side_II_died then 
			wesnoth.wml_actions.endlevel { result = "victory" }
		end 
	elseif game_mode == "duel" then
		if not boss_side_II_died then
			next_wave_turn = postwave_phase_init()
		end
	elseif game_mode == "solo" then 
		wesnoth.wml_actions.endlevel { result = "victory" }
	end
end


function raginature.boss_side_II_die()
	boss_side_II_died = true
	if game_mode == "coop"  then 
		if boss_side_I_died then 
			wesnoth.wml_actions.endlevel { result = "victory" }
		end 
	elseif game_mode == "duel" then
		if not boss_side_I_died then
			next_wave_turn = postwave_phase_init()
		end
	end
end



-- No hidding of the leader
function raginature.side_turn_end_event()
	for _, unit in ipairs(wesnoth.get_units { canrecruit = true, side = wesnoth.current.side }) do
		unit.status.uncovered = true
	end
end



function raginature.turn_refresh()
	local this_turn = wesnoth.current.turn
	local this_side = wesnoth.current.side
	
	if next_wave_turn and (this_turn == next_wave_turn+1) then
		current_wave_nb = current_wave_nb + 1
		next_wave_turn = get_next_wave_turn()
		if next_wave_turn then
				set_label_nextwave(next_wave_turn)
		end
		
	elseif this_turn == next_wave_turn then
	
		if  this_side == 1 then
			if not postwave_phase then
				create_wave_by_nb(current_wave_nb+1)
			else 
				local strength = postwave_strength_start + current_wave_nb * postwave_strength_increase
				create_wave(postwave, strength)
			end
			wesnoth.play_sound "foghorn-[1~3].ogg"

		elseif (this_side == 2) or (this_side == 4) then
			make_real_mobs(this_side)
		end
	end
end



-- >>
